
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include "lgc.h"
#include "ansi.h"
#include "crc32.h"
#include "game.h"
#include "output.h"

void                     processMessage();
void                     writeServerInfo();
void                     writeChannels();
void                     writeClients(int chan = -1);
void                     getAutoMsgs(char *chan = NULL);
void                     paging(int b);
void                     msgStr(char *fmt,...);
void                     msgWrapStr(int start, char *s);
int                      clientColor(char *s);
void                     setNormal();
void                     setColor(int c);
void                     setBold();
void                     setBlinking();
int                      colorNum(char *s);
void                     _werase(WINDOW * w);
void                     DOSwriteByte(char c);
void                     DOSwriteString(char *s, int l);
void                     DOSendPacket();

static void              waitKey();

static int               pagingEnabled,
                         pagingLines;
static CRC32             crc;

void
processMessage()
{
  int                      x,
                           y;

  switch (msg.type) {
  case MSG_PING:
    x = sizeof(int);
    y = CMD_PING;
    write(fd, &x, sizeof(int));
    write(fd, &y, sizeof(int));
    break;
  case MSG_CONNECT:
    gameClient = msg.data[0];
    gameInProgress = msg.data[1];
    initialNodes = msg.data[2];
    nodeAddr = msg.data[3];
    strcpy(gameSetup, msg.data + 4);
    networkStart();
    networkEnd();
    break;
  case MSG_NICKNAME:
    msgStr("<C %d>**</C> <C %d>%s</C> is now known as <C %d>%s</C>\n", FG_STAR, clientColor(msg.data), msg.data, clientColor(msg.data + strlen(msg.data) + 1), msg.data + strlen(msg.data) + 1);
    break;
  case MSG_CONNREQ:
    msgStr("<C %d>**</C> <C %d>NETWORK CONNECTION</C> from <C %d>%s</C>\n", FG_STAR, FG_CAPS, clientColor(msg.data), msg.data);
    break;
  case MSG_CLIENTCONN:
    if (!strcasecmp(msg.data, msg.data + strlen(msg.data) + 1))
      msgStr("<C %d>**</C> <C %d>%s</C>'s network has been activated\n", FG_STAR, clientColor(msg.data), msg.data);
    else
      msgStr("<C %d>**</C> <C %d>%s</C> connected to <C %d>%s</C>'s network\n", FG_STAR, clientColor(msg.data), msg.data, clientColor(msg.data + strlen(msg.data) + 1), msg.data + strlen(msg.data) + 1);
    break;
  case MSG_CLIENTDISC:
    if (!strcasecmp(msg.data, msg.data + strlen(msg.data) + 1))
      msgStr("<C %d>**</C> <C %d>%s</C>'s network has been de-activated\n", FG_STAR, clientColor(msg.data), msg.data);
    else
      msgStr("<C %d>**</C> <C %d>%s</C> disconnected from <C %d>%s</C>'s network\n", FG_STAR, clientColor(msg.data), msg.data, clientColor(msg.data + strlen(msg.data) + 1), msg.data + strlen(msg.data) + 1);
    break;
  case MSG_CLIENTJOINED:
    msgStr("<C %d>**</C> <C %d>%s</C> joined this channel\n", FG_STAR, clientColor(msg.data), msg.data);
    break;
  case MSG_CLIENTLEFT:
    msgStr("<C %d>**</C> <C %d>%s</C> left this channel\n", FG_STAR, clientColor(msg.data), msg.data);
    break;
  case MSG_SIGNON:
    msgStr("<C %d>**</C> <C %d>%s</C> signed on\n", FG_STAR, clientColor(msg.data), msg.data);
    break;
  case MSG_SIGNOFF:
    msgStr("<C %d>**</C> <C %d>%s</C> signed off\n", FG_STAR, clientColor(msg.data), msg.data);
    break;
  case MSG_ACTION:
    msgStr("<C %d>*</C> <C %d>%s</C> ", FG_STAR, clientColor(msg.data), msg.data);
    msgWrapStr(strlen(msg.data) + 3, msg.data + strlen(msg.data) + 1);
    break;
  case MSG_AUTOMSG:
    msgStr("<B>AutoMsg</B> from <C %d>%s</C> -> ", clientColor(msg.data), msg.data);
    msgWrapStr(strlen(msg.data) + 17, msg.data + strlen(msg.data) + 1);
    break;
  case MSG_DONEAUTOMSGS:
    paging(0);
    break;
  case MSG_BROADCAST:
    msgStr("<C %d>%s</C>: ", clientColor(msg.data), msg.data);
    msgWrapStr(strlen(msg.data) + 2, msg.data + strlen(msg.data) + 1);
    break;
  case MSG_MSGSENT:
    msgStr("-> <C %d>*</C><C %d>%s</C><C %d>*</C> ", FG_STAR, clientColor(msg.data), msg.data, FG_STAR);
    msgWrapStr(strlen(msg.data) + 6, msg.data + strlen(msg.data) + 1);
    break;
  case MSG_PRIVMSG:
    msgStr("<C %d>*</C><C %d>%s</C><C %d>*</C> -> ", FG_STAR, clientColor(msg.data), msg.data, FG_STAR);
    msgWrapStr(strlen(msg.data) + 6, msg.data + strlen(msg.data) + 1);
    break;
  }
  if (!optDOS) {
    getyx(cmdW, y, x);
    wmove(cmdW, y, x);
    wrefresh(cmdW);
  }
}

void
writeServerInfo()
{
  refreshServerInfo();
  msgStr("\n");
  msgStr("  Server start time:  <C BLACK><B>%s</C>", ctime(&serverInfo.startTime));
  msgStr("      Server uptime:  <C BLACK><B>%.2f days</C>\n", (float) serverInfo.upTime / (24 * 60 * 60));
  msgStr("    Nodes supported:  <C BLACK><B>%d</C>\n", serverInfo.maxNet);
  msgStr("    Channels formed:  <C BLACK><B>%d</C>\n", serverInfo.nChannels);
  msgStr("  Clients connected:  <C BLACK><B>%d</C>\n", serverInfo.nClients);
  msgStr("Network connections:  <C BLACK><B>%d active / %d total</C>\n", serverInfo.activeNetConnections, serverInfo.netConnections);
  msgStr("       Active games:  <C BLACK><B>%d</C>\n", serverInfo.activeGames);
  msgStr("\n");
}

void
writeChannels()
{
  int                      i,
                           j;

  refreshChannels();
  for (j = 0, i = 0; i < nChannels; i++)
    j += channel[i].nConnects;
  j = serverInfo.maxNet - j;
  msgStr("\n");
  paging(1);
  msgStr("<B>Channel        Con:Tot:Msg     Net-Nodes     Topic</B>\n");
  msgStr("\
-------------------------------------------------------------------------------\n");
  for (i = 0; i < nChannels; i++) {
    msgStr("#%-14s%3d:%3d:%3d      %3d/%-3d      %s\n", channel[i].name, channel[i].nConnects, channel[i].nClients, channel[i].nAutoMsgs, min(j, channel[i].maxNet), channel[i].maxNet, channel[i].topic);
  }
  msgStr("\n");
  getAutoMsgs();
}

void
writeClients(int chan)
{
  int                      i,
                           n;
  char                     connStr[64];

  if (chan < 0)
    chan = client[clientNo].channel;
  msgStr("\n");
  paging(1);
  for (n = 0, i = 0; i < nClients; i++)
    if (client[i].fd >= 0 && client[i].channel == chan)
      n++;
  if (!n) {
    msgStr("<B>**</B> No clients in #%s\n\n", channel[chan].name);
  } else {
    msgStr("<B>**</B> Listing of clients in #%s\n\n", channel[chan].name);
    msgStr("<B>Nickname      Connection    Name</B>\n");
    msgStr("\
-------------------------------------------------------------------------------\n");
    for (i = 0; i < nClients; i++) {
      if (client[i].channel != chan)
	continue;
      if (client[i].fd >= 0) {
	if (client[i].ccid >= 0) {
	  if (client[i].ccid == i)
	    sprintf(connStr, " <C %d>*</C> <C %d>SELF</C> <C %d>*</C>     ", FG_STAR, FG_CAPS, FG_STAR);
	  else {
	    int                      j;

	    sprintf(connStr, "<C %d>%s</C>", clientColor(client[client[i].ccid].name), client[client[i].ccid].name);
	    for (j = strlen(client[client[i].ccid].name); j < 14; j++)
	      strcat(connStr, " ");
	  }
	} else
	  strcpy(connStr, " - none -     ");
	msgStr("<C %d>%-14s</C>%s%s\n", clientColor(client[i].name), client[i].name, connStr, client[i].gecos);
      }
    }
    msgStr("\n");
  }
  getAutoMsgs(channel[chan].name);
}

void
getAutoMsgs(char *chan)
{
  int                      cmdLen = sizeof(int);
  char                     cmd[64];

  *(int *) cmd = CMD_GETAUTOMSGS;
  if (chan) {
    strcpy(cmd + sizeof(int), chan);

    cmdLen += strlen(chan) + 1;
  } else {
    strcpy(cmd + sizeof(int), ".");

    cmdLen += 2;
  }
  request(&cmd, cmdLen);
}

void
paging(int b)
{
  if (b) {
    if (optDOS) {
      DOSwriteByte(PKT_PSTART);
      DOSendPacket();
    } else {
      pagingEnabled = 1;
      pagingLines = 0;
    }
  } else {
    if (optDOS) {
      DOSwriteByte(PKT_PEND);
      DOSendPacket();
    } else {
      pagingEnabled = 0;
    }
  }
}

void
msgStr(char *fmt,...)
{
  char                     fmtCmd[16],
                           buf[1024];
  int                      nPrinted = 0,
                           i,
                           j,
                           l,
                           neg;
  va_list                  ap;

  va_start(ap, fmt);
  vsprintf(buf, fmt, ap);
  i = 0;
  l = strlen(buf);
  if (optDOS)
    DOSwriteByte(PKT_MSG);
  while (i < l) {
    if (buf[i] == '\\') {
      if (buf[i + 1] == '\\') {
	if (optDOS)
	  DOSwriteByte('\\');
	else
	  waddch(msgW, '\\');
      } else if (buf[i + 1] == '<') {
	if (optDOS)
	  DOSwriteByte('<');
	else
	  waddch(msgW, '<');
      } else {
	if (optDOS)
	  DOSwriteByte(buf[i + 1]);
	else
	  waddch(msgW, buf[i + 1]);
      }
      i += 2;
    } else if (buf[i] == '<') {
      i++;
      if (buf[i] == '/') {
	neg = 1;
	i++;
      } else
	neg = 0;
      for (j = 0; isalpha(buf[i]); i++)
	fmtCmd[j++] = tolower(buf[i]);
      fmtCmd[j] = '\0';
      if (!strncmp(fmtCmd, "b", 1)) {
	if (neg)
	  setColor(COLOR_WHITE);
	else
	  setBold();
      } else if (!strncmp(fmtCmd, "c", 1)) {
	if (neg)
	  setColor(COLOR_WHITE);
	else {
	  for (i++; i < l && !isalpha(buf[i]) && !isdigit(buf[i]); i++);
	  if (i == l)
	    continue;
	  for (j = 0; isalpha(buf[i]) || isdigit(buf[i]); i++)
	    fmtCmd[j++] = buf[i];
	  fmtCmd[j] = '\0';
	  if (isdigit(fmtCmd[0]))
	    setColor(atoi(fmtCmd));
	  else
	    setColor(colorNum(fmtCmd));
	}
      } else if (!strncmp(fmtCmd, "t", 1)) {
	if (neg)
	  setColor(COLOR_WHITE);
	else
	  setColor(FG_TOPIC);
      }
      while (i < l && buf[i++] != '>');
    } else {
      if (buf[i] == '\n' && !optDOS)
	wrefresh(msgW);
      if (buf[i] != '\n' || nPrinted < COLS) {
	if (buf[i] == '\n' || nPrinted + 1 == COLS) {
	  if (pagingEnabled && ++pagingLines == LINES - 1) {
	    waitKey();
	    pagingLines = 0;
	  }
	  nPrinted = 0;
	} else
	  nPrinted++;
	if (optDOS)
	  DOSwriteByte(buf[i]);
	else
	  waddch(msgW, buf[i]);
	if (buf[i] == '\n')
	  nPrinted = 0;
      }
      i++;
    }
  }
  if (optDOS)
    DOSendPacket();
  else
    wrefresh(msgW);
  va_end(ap);
}

void
msgWrapStr(int start, char *s)
{
  int                      i = 0,
                           j,
                           l = strlen(s);
  char                     c1,
                           c2;

  while (i < l) {
    if (l - i + 1 < COLS - start) {
      msgStr("%s\n", s + i);
      break;
    } else {
      j = COLS - start - 1;
      while (j && s[i + j] != ' ')
	j--;
      if (!j) {
	j = COLS - start - 1;
	c1 = s[i + j - 1];
	c2 = s[i + j];
	s[i + j - 1] = '-';
	s[i + j] = '\0';
	msgStr("%s\n", s + i);
	s[i + j - 1] = c1;
	s[i + j] = c2;
	i += j - 1;
      } else {
	s[i + j] = '\0';
	msgStr("%s\n", s + i);
	s[i + j] = ' ';
	i += j;
	while (i < l && s[++i] == ' ');
      }
      msgStr("<B>+</B>");
      start = 1;
    }
  }
}

int
clientColor(char *s)
{
  CRC32                    crc;
  int                      l = strlen(s);

  for (int i = 0; i < l; i++)
    crc.addByte(s[i]);
  int                      color = crc.CRC() % 16;

  if (!color || color == 7)
    color++;
  if (color >= 8) {
    color -= 8;
    color |= A_BOLD;
  }
  return color;
}

void
setNormal()
{
  if (optDOS)
    ansiColor(ANSI_NORMALATTR);
  else
    wattr_set(msgW, A_NORMAL);
}

void
setColor(int c)
{
  int                      bold = c & A_BOLD;

  if (bold)
    c &= ~A_BOLD;
  if (optDOS)
    ansiColor(c + 30);
  else
    wattr_set(msgW, COLOR_PAIR(c + 2));
  if (bold)
    setBold();
}

void
setBold()
{
  if (optDOS)
    ansiAttr(ANSI_BOLD);
  else
    wattr_on(msgW, A_BOLD);
}

void
setBlinking()
{
  if (optDOS)
    ansiAttr(ANSI_BLINKING);
  else
    wattr_on(msgW, A_BLINK);
}

int
colorNum(char *s)
{
  if (!strncasecmp(s, "BLACK", 3))
    return COLOR_BLACK;
  else if (!strncasecmp(s, "RED", 1))
    return COLOR_RED;
  else if (!strncasecmp(s, "GREEN", 1))
    return COLOR_GREEN;
  else if (!strncasecmp(s, "YELLOW", 1))
    return COLOR_YELLOW;
  else if (!strncasecmp(s, "BLUE", 3))
    return COLOR_BLUE;
  else if (!strncasecmp(s, "MAGENTA", 1))
    return COLOR_MAGENTA;
  else if (!strncasecmp(s, "CYAN", 1))
    return COLOR_CYAN;
  else if (!strncasecmp(s, "WHITE", 1))
    return COLOR_WHITE;
  else
    return COLOR_WHITE;
}

void
_werase(WINDOW * win)
{
  int                      x,
                           y;

  for (y = 0; y <= win->_maxy; y++) {
    wmove(win, y, 0);
    for (x = 0; x <= win->_maxx; x++)
      waddch(win, ' ');
  }
  wmove(win, 0, 0);
}

void
DOSwriteByte(char c)
{
  putchar(c);
  if ((byte) c == 0xff)
    putchar(c);
  ::crc.addByte(c);
}

void
DOSwriteString(char *s, int l)
{
  for (int i = 0; i < l; i++)
    DOSwriteByte(s[i]);
}

void
DOSendPacket()
{
  word                     crc = ::crc.CRC();
  char                    *s = (char *) &crc;

  for (int i = 0; i < (int) sizeof(word); i++)
    DOSwriteByte(s[i]);
  putchar(0xff);
  putchar(0);
  ::crc.init();
}

void
waitKey()
{
  _werase(cmdW);
  wrefresh(cmdW);
  mvwaddstr(cmdW, 0, (COLS - 33) / 2, ">>> Press any key to continue <<<");
  wrefresh(cmdW);
  wgetch(cmdW);
  _werase(cmdW);
  mvwchgat(cmdW, 0, 0, -1, A_REVERSE, COLOR_PAIR(1), NULL);
  wrefresh(cmdW);
}
